"""
The `records list` command can be used to list recorded files within the registry.
By default this will list all recorded files in the registry.  The results can be restricted
using the ``-f`` and ``-e`` flags.

Usage
-----

    >>> records list [-f <flag>] [-e <pattern>]

    where ``<flag>``, is a flag to search for and ``<pattern>``, is a regular expression to search for.
    Both ``<flag>`` and ``<pattern>`` *can* be specified at the same time. 

    Note
    ----
    The ``-f`` flag can only be used to search for a single flag. To search for multiple flags at a time, define a flag group first and then search for it's label using 'group:your_group'.


"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "List file records."
    parser = parent.add_parser( "list", description = descr, help = descr )
    parser.add_argument( "-f", "--flag", help = "The flag search for. Note, this may only be a single flag! To search for multiple flags at a time, define a flag group first and then search for it's label using 'group:your_group'.", default = None )
    parser.add_argument( "-e", "--pattern", help = "The regular expression to search for.", default = None )
    parser.set_defaults( func = search )

def search( args ):
    """
    The core function to search for entries.
    """
    import filerecords.api as api
    import filerecords.api.utils as utils

    logger = utils.log()
    reg = api.Registry( "." )

    records = reg.search( args.pattern, flag = args.flag )

    if len( records ) == 0:
        logger.info( "No records found." )
        return
    
    output = ""
    for record in records:
        output += f"{record.relpath[3:]}  ({', '.join(record.flags)})\n"

    print( output.rstrip() )


"""
The `records export` command can be used export a YAML or markdown summary (manifest) of all 
records in the registry.

Usage:

    >>> records export <yaml|md|both> -f <filename>

    where either ``yaml``, ``md``, or ``both`` can be specified to export the records to a yaml or markdown file, or both.
    The ``-f`` option can be used to specify a specific filename to export to. 
    Note, `.yaml` and `.md` are added to the filename automatically.
"""

def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Export the registry to a file manifest."
    parser = parent.add_parser( "export", description = descr, help = descr )
    parser.add_argument( "format", help = "The export format, which can be either yaml, markdown, or both.", choices = ["md", "yaml", "both"] )
    parser.add_argument( "-f", "--filename", help = "The filename to export to. If not specified, a default 'registry-{timestamp}' file will be created.", default = None )
    parser.set_defaults( func = export )


def export( args ): 
    """
    The core function to export the registry.
    """
    import filerecords.api as api
    import filerecords.api.settings as settings
    from datetime import datetime
    # from filerecords.api.utils import log
 
    # logger = log()
    reg = api.Registry( "." )

    if args.filename is None:
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        args.filename = f"{settings.registry_export_name}-{timestamp}"

    if args.format == "yaml" or args.format == "both":
        reg.to_yaml( timestamp = True, filename = args.filename + ".yaml" )

    if args.format == "md" or args.format == "both":
        reg.to_markdown( timestamp = True, filename = args.filename + ".md" )

    # logger.info( f"Exported registry to {args.filename}" )
    print( f"Exported registry to {args.filename}" )


"""
The `records undo` command can be used to remove a file or directory's last added comment.

Usage
-----

    >>> records undo [-f <flags>] <filename>

    where ``<filename>`` is the path to the file whose last comment to remove. 
    If left blank the comment is removed from the registry itself.
    ``-f <flags>``, removes any number of flags. These can include defined flag group labels.

    Note
    ----
    When flags are specified, then only the flags are removed while the comments are left untouched.
"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Remove flags or the latest comment from a file or directory."
    parser = parent.add_parser( "undo", description = descr, help=descr )
    parser.add_argument( "filename", nargs = "?", help = "The file whose metadata to undo. If left blank the actions are applied to the registry itself", default = None )
    parser.add_argument( "-f", "--flags", help = "Any flags to remove.", nargs="+", default = None )
    parser.set_defaults( func = undo )

def undo( args ):
    """
    The core function to undo comments or files.
    """
    import filerecords.api as api
    import filerecords.api.utils as utils

    logger = utils.log()

    reg = api.Registry( "." )
    
    if args.flags and len(args.flags) == 1:
        args.flags = args.flags[0]
    
    if not args.filename:

        if args.flags:
            reg.remove_flags( args.flags )
        else:
            reg.undo_comment()

        reg.save()

    else:

        record = reg.get_record( args.filename )
        if record is None:
            logger.error( f"No record for {args.filename}" )
            return
        
        if args.flags:
            record.remove_flags( args.flags )
        else:
            record.undo_comment()
        
        logger.info( f"Updated {args.filename} in the registry." )
        record.save()


"""
`Registry` is the main class handling the `filerecords` registry.
The registry itself is a hidden directory containing the records of files and directories in yaml format.


API Usage
=========

In the code API, a registry can be accessed by setting up a Registry object, which will automatically locate the 
closest available registry, but can also initialize a new registry. 

.. code-block:: python

    from filerecords.api import Registry

    # Call a registry from the current directory.
    reg = Registry()


Initializing new registries
---------------------------

The above code will try to find a registry in any directory that higher in the directory tree than the current directory.
If none is found, a new registry is automatically initialized in the current directory.

In case a registry is found, the registry can be accessed now. However, a new registry can still be initialized in the current directory
by calling `reg.init()`.

.. code-block:: python

    # Initialize a new registry in the current directory 
    # (even if a registry was found in a parent directory).
    reg.init()


Accessing the registry
----------------------

The registry's own metadata can be accessed via the `comments` and `flags` attributes directly.

.. code-block:: python

    # Get the registry's comments.
    reg.comments

    # Get the registry's flags.
    reg.flags

    # Get the defined flag groups.
    reg.groups


New comments and flags can be added to the registry itself using the `add_comment()` and `add_flags()` methods.

.. code-block:: python

    # Add a new comment.
    reg.add_comment( "This is a new comment." )

    # Add a new flag.
    reg.add_flag( "new_flag" )  


After editing the comments or flags of a registry, the changes must be saved using the `save()` method.

.. code-block:: python

    # Save the registry.
    reg.save()


To access records from the registry the `get_record()` method can be used.

.. code-block:: python

    # Get a record from the registry.
    record = reg.get_record( "path/to/file" )

    # e.g. 
    record = reg.get_record( "results/gsea_20082022.tsv" )


This will return a `FileRecord` object which now allows to access the metadata of the file.

Alternatively, if the precise filename is not known or a number of files shall be found, the `search()` 
method can be used to find records based on their flags or on regex patterns in their filenames.

.. code-block:: python

    # Search for records with the "important" flag.
    records = reg.search( flag = "important" )

    # Search for records with the "important" flag and which are bam files.
    records = reg.search( flag = "important", pattern = ".*\.bam" )


.. note::

    Search always applies *and* logic to to the pattern and flag. Only a single flag is supported for searching.
    To include multiple flags in a search, first define a flag group and then search for the group's label flag.

    For instance, assume we wish to search for flags "important" and "results" at the same time.
    We first create a group (see below) containing these flags and then search for the group. Then we need to flag our files of interest
    with the group. This is not an ideal system yet and may be updated in the future.


Flags and flag groups
---------------------

Flags are used to mark files in the registry. They can be used to mark files as important, as results, as temporary files, etc.
They are a shorthand for the details in the comments. Often it is desirable to add multiple flags at the same time to a file. To 
avoid having to repeatedly add multiple flags at the same time, registries support `flag groups` which summarize multiple flags by one label.

.. code-block:: python

    # Add a new flag group
    reg.add_group( "supergroup", ["important", "results"] )

    # now add a new file and flag it as member of the supergroup
    reg.add( "my_superfile.txt", comment = "superfile", flags = "supergroup" )

    # search for all files flagged as supergroup
    records = reg.search( flag = "group:supergroup" )


Adding and editing records
--------------------------

New files can be added to the registry using the `add()` method. 
If a file is already recorded, the `update()` method must be used instead.

.. code-block:: python

    # add a "results/" directory to the registry
    reg.add( "results/", comment = "the main results", flags = ["results", "important"] )

    # to now add additional comments or flags to the "results/" directory use update instead
    reg.update( "results/", comment = "an additional comment about the results", flags = ["perhaps_another_flag"] )


Records (i.e. files or directories) can be `moved` and/or `removed` from the registry. By default these actions also affect the files in the filesystem.
I.e. a file is by default also deleted if it is removed from the registry - this can be controlled, however.

.. code-block:: python

    # Move a file to a new location.
    record.move( "/path/to/file", "new/path/to/file" )

    # Only adjust the references in the registry 
    # (maybe because you already moved the file and forgot to update the registry).
    record.move( "/path/to/file", "new/path/to/file", keep_file = True )

    # Remove a record from the registry.
    record.remove( "/path/to/file" )

    # Only remove the references in the registry but keep the file in the filesystem.
    record.remove( "/path/to/file", keep_file = True )


Exporting the contents of the registry
--------------------------------------

The registry contents can be represented in markdown format using the `to_markdown()` method, 
or in yaml format using the `to_yaml()` method.

.. code-block:: python

    # Export the registry to a markdown file.
    reg.to_markdown( timestamp = True, filename = "registry.md" )

    # Export the registry to a yaml file.
    reg.to_yaml( timestamp = True, filename = "registry.yaml" )

"""

from datetime import datetime
import shutil
import os
import subprocess
import pandas as pd


import filerecords.api.base as base
import filerecords.api.file_record as file
import filerecords.api.utils as utils
import filerecords.api.settings as settings

logger = utils.log()

class Registry(base.BaseRecord):
    """
    The main class of a filrecords registry. It loads the registry 
    information from a parent directory and makes the data accessible.

    Note
    
        This class will search for existing registries automatically in the filesystem and 
        initialize a new registry if none are found. However, even if one is found a new registry can be
        initialized in the current directory by calling `init()`.

    Parameters
    ----------
    directory : str
        The directory to load the registry from. 
        By default the current working directory is used. 
    """
    def __init__(self, directory : str = "." ):
        super().__init__()

        # self.directory = os.path.abspath( directory ) 
        self.directory = subprocess.check_output( f"cd {directory} ; pwd", shell = True ).decode().strip()
        
        self.index = None

        self._initialized = False
        self._find_registry()
        
        if not self._initialized:
            self._load_registry()

    def init( self, permissions : (int or str) = None ):
        """
        Initialize a new registry in the given directory.

        Parameters
        ----------
        permissions : int or str
            The permissions to use for the registry directory. 
            By default the permissions of the parent directory are used.
        """
        self._initialized = True
        utils.make_new_registry( self.directory, perms = permissions )

        self.registry_dir = os.path.join( self.directory, settings.registry_dir ) 
        self._load_registry()

    def save( self ):
        """
        Save the registry state and updated metadata.
        """
        self.index.to_csv( self.indexfile, index = False, sep = "\t" )
        super().save()

    
    def get_record( self, filename : str ):
        """
        Get the record of a file in the registry.

        Parameters
        ----------
        filename : str
            The filename of the file to get the record of.

        Returns
        -------
        FileRecord or list
            The record of the file or a list of records.
        """
        logger.debug( f"directory={self.directory}")
        logger.debug( f"registry_dir={self.registry_dir}")
        filename = os.path.join( self.directory, filename )
        logger.debug( f"filename={filename}" )
        match = os.path.relpath( filename, self.registry_dir )
        logger.debug( f"match={match}" )

        match = self.index.relpath == match 
        logger.debug( f"match={match}" )

        if match.any():

            if match.sum() > 1:
                logger.warning( f"More than one record found for {filename}." )
                return [ file.FileRecord( registry = self, id = i ) for i in self.index.loc[match, "id"].values ] 
            
            id = self.index.loc[match, "id"].values[0]
            return file.FileRecord( registry = self, id = id )

        return None

    def add_group( self, label : str, flags : list ):
        """
        Add a flag group to the registry.


        Note
        ----
        This will not automatically save the registry's metadata,
        use the save() method to do so.

        Parameters
        ----------
        label : str
            The label of the group.
        flags : list
            The flags of the group.
        """
        if not isinstance( flags, list ):
            flags = [ flags ]
        flags += [ f"group:{label}" ]

        self.metadata["groups"].update( { label : flags  } )
        self.add_flags( flags )

    def add( self, filename : str, comment : str = None, flags : list = None ):
        """
        Add a new file to the registry.

        Parameters
        ----------
        filename : str 
            The filename of the file to add.
        comment : str
            The comment to add to the file.
        flags : list
            Any flags to add. This can also be a defined flag-group label.
        """
        if not os.path.exists( filename ):
            raise FileNotFoundError( f"File {filename} does not exist. Can only comment existing files..." )
        
        if not comment and not flags:
            raise ValueError( "No flags or comment given. At least one must be given." )


        record = file.FileRecord( registry = self, filename = filename )
        new_id = record.id

        logger.debug( f"adding: {filename} (filename = { os.path.basename(record.relpath) })" )
        
        if comment:
            record.add_comment( comment )
    
        if flags:
            record.add_flags( flags )
        
        index_entry = pd.DataFrame( { "id" : [new_id], "filename" : [os.path.basename(record.relpath)], "relpath" : [record.relpath] } )
        self.index = pd.concat( [self.index, index_entry], ignore_index = True )

        record.save()
        # logger.info( f"Added {filename} to the registry." )


    def update( self, filename : str, comment : str = None, flags : (str or list) = None ):
        """
        Update an existing file record.

        Parameters
        ----------
        filename : str
            The filename of the file to update.
        comment : str
            The new comment to add to the file.
        flags : str or list
            The new flags to add to the file.
        """
        record = self.get_record( filename )

        if record is None:
            logger.warning( f"No record found for {filename}." )
            return

        elif isinstance( record, list ):
            logger.warning( f"More than one record found for {filename}, can only edit one record at a time." )
            return

        if comment:
            record.add_comment( comment )
        if flags:
            record.add_flags( flags )

        record.save()
        # logger.info( f"Updated {filename} in the registry." )

    def search( self, pattern: str = None, flag : str = None ):
        """
        Search for records in the registry either through a filename pattern or by a flag.

        Parameters
        ----------
        pattern : str
            The filename pattern to search for.
        flag : str
            The flag to search for. Note, this can only be a single flag!
            To search for multiple flags, first define a flag-group and then search
            for the group label using `group:yourgroup`. 
        
        Returns
        -------
        list
            A list of FileRecord objects of record entries matching the search criteria.
        """
        records = None

        if pattern:
            records = self.index["filename"].str.contains( pattern, regex = True )
            if records.sum() == 0:
                logger.warning( f"No records found for pattern {pattern}." )
            else:
                records = self.index.loc[records, "id"].values
                records = [ file.FileRecord( self, id = id ) for id in records ]
        
        if flag:
            if not records:
                records = [ file.FileRecord( self, id = id ) for id in self.index.id ]
            records = [ i for i in records if flag in i.flags ]
        
        if not pattern and not flag:
            logger.warning( "No search criteria specified, returning all records." )
            records = [ file.FileRecord( self, id = id ) for id in self.index.id ]

        return records

    def move( self, current : str, new : str, keep_file : bool = False ):
        """
        Move a file to a new location.

        Parameters
        ----------
        current : str
            The filename of the file to move.
        new : str
            The new filename to move the file to.
        keep_file : bool
            If True only the path reference is adjusted within the registry.
            If False the file moving will also be performed.
        """
        record = self.get_record( current )

        if record is None:
            logger.warning( f"No record found for {current}." )
            return

        elif isinstance( record, list ):
            logger.warning( f"More than one record found for {current}, can only edit one record at a time." )
            return

        # Note: the {}_path are relative to the registry dir and therefore
        # registry internal, while the actual current and new are relative to the users
        # current working directory and must not be altered for file moving...

        current_path = os.path.relpath( current, self.registry_dir )
        new_path = os.path.relpath( new, self.registry_dir )
        mask = self.index.relpath == current_path
        
        self.index.loc[ mask, "relpath" ] = new_path
        self.index.loc[ mask, "filename" ] = os.path.basename( new )

        if not keep_file:
            os.rename( current, new )

        self.save()

    def remove( self, filename : str, keep_file : bool = False ):
        """
        Remove a file from the registry.

        Parameters
        ----------
        filename : str
            The filename of the file to remove.
        keep_file : bool
            If True, the file will not be removed from the filesystem, only its records in the registry.
        """
        record = self.get_record( filename )
        if record is None:
            logger.warning( f"No record found for {filename}." )
            return

        if not keep_file:
            if os.path.isfile( filename ):
                os.remove( filename )
            elif os.path.isdir( filename ):
                shutil.rmtree( filename )
            else:
                logger.warning( f"{filename} is not a file or directory. Cannot remove." )

        os.remove( record.metafile )
        self.index = self.index[ self.index.id != record.id ]
        
        self.save()
        # logger.info( f"Removed {filename} from the registry." )


    def to_yaml( self, include_records : bool = True, timestamp : bool = False, filename : str = None ):
        """
        Convert the source registry to a single YAML file.

        Parameters
        ----------
        include_records : bool
            Include the records in the markdown.
        timestamp : bool
            Add a timestamp in the markdown.
        filename : str (optional)
            The filename of the yaml file to create.
            If none is provided, no file is created.

        Returns
        -------
        dict
            The assembled dictionary of the registry.
        """

        _dict = dict( self.metadata )
        _dict["directory"] = self.directory
        if timestamp:
            _dict["timestamp"] = datetime.now().strftime( "%Y-%m-%d %H:%M:%S" )

        if include_records:
            records = [ file.FileRecord( self, id = id ) for id in self.index.id ]
            records = { record.relpath[3:] : record.to_yaml() for record in records }
            _dict[ "records" ] = records

        if filename is not None:
            utils.save_yamlfile( filename, _dict )

        return _dict

    def to_markdown( self, include_records : bool = True, timestamp : bool = False, filename : str = None ):
        """
        Convert the metadata to a markdown representation.

        Parameters
        ----------
        include_records : bool
            Include the records in the markdown.
        timestamp : bool
            Add a timestamp in the markdown.
        filename : str (optional)
            The filename of the markdown file to create.
            If none is provided, no file is created.

        Returns
        -------
        str
            The markdown representation of the registry.
        """
        # add basic information and timestamp of manifest creation
        text = f"# {self.directory}\n\n"
        if timestamp: 
            text += f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n\n"
        
        # add registry's own comments
        text += "## Registry comments\n\n"
        for timestamp in self.comments:
            comment, user = list( self.comments[timestamp].values() )
            text += f"{settings.comment_format( comment, user, timestamp)}\n\n"

        # add all flags and flag groups
        text += "## Registered flags\n\n"
        for i in sorted( self.flags ):
            text += f"- {i}\n"
        text += "\n"

        text += "### Flag groups\n\n"
        text += "| Group | Flags |\n"
        text += "|------|------|\n"
        for label, flags in self.groups.items():
            text += f"| {label} | {', '.join(flags)} |\n"

        if include_records:
            text += "\n"
            text += "## Records\n\n"
            if len( self.index.id ) == 0:
                text += "No records found."
            else:
                for record in self.index.id:
                    record = file.FileRecord( self, id = record )
                    text += record.to_markdown( comments_header = False )

        if filename is not None:
            with open( filename, "w" ) as f:
                f.write( text )

        return text

    def base_has_registry( self ):
        """
        Checks if the current directory already has a registry.
        """
        return os.path.exists( os.path.join( self.directory, settings.registry_dir ) )


    @property
    def groups( self ) -> dict:
        """
        Get the defined flag-groups.
        """
        return self.metadata["groups"]

    def _find_registry( self ):
        """
        Finds the local registry associated with the given directory.
        This will parse the directory hierarchy upwards until it finds a registry.
        If none are found it will initialize a new registry in the given directory.
        """
        self.registry_dir = utils.find_registry( self.directory )

        if self.registry_dir is None:
            logger.info( "No local registry found. Initializing a new registry." )
            self.init()
            self.registry_dir = os.path.join( self.directory, settings.registry_dir )
        
        
    def _load_registry( self ):
        """
        Loads the registry data from the indexfile and metafile
        """
        self.indexfile = utils.get_indexfile( self.registry_dir )
        self.metafile = utils.get_metafile( self.registry_dir )

        self.index = utils.load_indexfile( self.indexfile )
        self.metadata = utils.load_yamlfile( self.metafile )

    def __repr__( self ):
        return f"{self.__class__.__name__}(directory = {self.directory}, registry_in = {os.path.dirname( os.path.dirname( self.registry_dir ) ) })"


"""
The `records ls` command can be used to list files from the current directory for which records are available.
The results can be restricted using flags or regex patterns.

Usage
-----

    >>> records ls [-f <flag>] [-e <pattern>]

    where ``<flag>``, is a flag to search for and ``<pattern>``, is a regular expression to search for.
    Both ``<flag>`` and ``<pattern>`` *can* be specified at the same time. 

    Note
    ----
    The ``-f`` flag can only be used to search for a single flag. To search for multiple flags at a time, define a flag group first and then search for it's label using 'group:your_group'.

"""

def setup( parent ):
    """
    Set up the CLI
    """
    descr = "List files in the current directory that have records."
    parser = parent.add_parser( "ls", description = descr, help = descr )
    parser.add_argument( "-f", "--flag", help = "The flag search for. Note, this may only be a single flag! To search for multiple flags at a time, define a flag group first and then search for it's label using 'group:your_group'.", default = None )
    parser.add_argument( "-e", "--pattern", help = "The regular expression to search for.", default = None )
    parser.set_defaults( func = search )

def search( args ):
    """
    The core function to search for entries in the local directory.
    """
    
    import os, subprocess
    import filerecords.api as api
    import filerecords.api.utils as utils

    logger = utils.log()

    reg = api.Registry( "." )
        
    records = reg.search( args.pattern, flag = args.flag )

    # now restrict to those records that are stored in the current directory.
    # The current directory can either be specified by a path or be the base directory anway
    # in which case the relpath is empty.
    current = subprocess.check_output( "pwd", shell = True ).decode() # os.getcwd() keeps resolving symbolic links and without shell the same here...
    current = os.path.relpath( current.strip(), reg.registry_dir )
    logger.debug( f"{current=}")
    records = [ i for i in records if os.path.dirname( i.relpath ) == current ]

    if len( records ) == 0:
        logger.info( "No records found." )
        return

    output = ""
    for record in records:
        output += f"{record.relpath[3:]}  ({', '.join(record.flags)})\n"

    print( output.rstrip() )


"""
These are the default settings that control filerecords behavior.
"""
import logging

# ----------------------------------------------------------------
#   General settings
# ----------------------------------------------------------------

log_level = logging.INFO
"""The default logging level"""

# ----------------------------------------------------------------
#   Formatting settings
# ----------------------------------------------------------------
comment_format = lambda comment, name, timestamp : f"{comment}  |  {name} @ {timestamp.strftime('%Y-%m-%d %H:%M:%S')}"
"""The default format in which comments shall be represented in markdown files. This must be a function accepting a `comment`, `name`, and `timestamp (datetime object)` argument and return a string."""

# ----------------------------------------------------------------
#   Basic files
# ----------------------------------------------------------------

registry_dir = ".registry"
"""The directory to store the registry in"""

indexfile = "INDEXFILE"
"""The name of the indexfile, which will map filenames (basenames) to unique identifiers within the registry."""

registry_metafile = "METAFILE"
"""The name of the file storing the registry's own metadata - i.e. registry comments and the associated flags and flag groups."""

registry_export_name = "registry"
"""The default name of exported registry file(s) in yaml or markdown format"""

# ----------------------------------------------------------------
#   File architecture
# ----------------------------------------------------------------

indexfile_header = "id\tfilename\trelpath\n"
"""The header of the indexfile"""

entryfile_template = {
                        "comments" : {},
                        "flags" : [],
                    }
"""The template for file record entries"""


"""
This is the base class for the `filerecords` API classes.

.. note::

    This is not intended to be used directly.
    
"""

from datetime import datetime
import os

import filerecords.api.utils as utils

logger = utils.log()

class BaseRecord:
    """
    The basic record handling class for storing metadata.

    Parameters
    ----------
    filename : str
        The yaml file storing the record metadata.
    metadata : dict
        The metadata of the record.
    """
    def __init__(self, filename : str = None, metadata : dict = None ):
        self.metafile = filename
        self.metadata = metadata

        if self.metafile:
            self.load( self.metafile )

    def save( self ):
        """
        Save the metadata.
        """
        utils.save_yamlfile( self.metafile, self.metadata )

    def load( self, filename : str ):
        """
        Load yaml metadata from a file.
        
        Parameters
        ----------
        filename : str
            The path to the yaml file.
        """
        self.metafile = filename 
        self.metadata = utils.load_yamlfile( filename )

    def lookup_last( self ) -> dict:
        """
        Get the last comment.

        Returns
        -------
        dict
            The last comment dictionary with timestamp as key, 
            user and comment as values.
        """
        if len( self.metadata["comments"] ) == 0:
            logger.info("No comments found.")
            return None

        last = sorted( list( self.comments.keys() ) )[-1]
        return { last : self.metadata["comments"].get( last ) }

    def undo_comment( self ):
        """
        Undo the last comment.

        Note
        ----
        This will not automatically save the metadata,
        use the save() method to do so.
        """
        last = sorted( list( self.comments.keys() ) )[-1]
        self.metadata["comments"].pop( last )


    def add_comment( self, comment : str ):
        """
        Add a comment to the registry.

        Note
        ----
        This will not automatically save the metadata,
        use the save() method to do so.

        Parameters
        ----------
        comment : str
            The comment to add.
        """
        user = os.environ["USER"] 
        self.metadata["comments"].update(  { datetime.now() : {
                                                                 "comment" : comment, 
                                                                 "user" : user 
                                                            } 
                                            }  )

    def remove_flags( self, flag : (str or list) ):
        """
        Remove one or more flags from the registry.

        Note
        ----
        This will not automatically save the metadata,
        use the save() method to do so.

        Parameters
        ----------
        flag : str or list
            The flag(s) to remove.
        """
        if not isinstance( flag, list ):
            flag = [ flag ]
        logger.debug( "Removing flags: {}".format( flag ) )

        for f in flag:
            self.metadata["flags"].remove( f )

    def add_flags( self, flag : (str or list) ):
        """
        Add a new flag to the registry.

        Note
        ----
        This will not automatically save the metadata,
        use the save() method to do so.

        Parameters
        ----------
        flag : str or list
            The flag(s) to add.
        """
        if not isinstance( flag, list ):
            flag = [ flag ]
        logger.debug( "Adding flags: {}".format( flag ) )

        self.metadata["flags"] += flag
        logger.debug( "(before set) metadata['flags']: {}".format( self.metadata["flags"] ) )

        self.metadata["flags"] = list( set( self.metadata["flags"] ) )
        logger.debug( "(after set) metadata['flags']: {}".format( self.metadata["flags"] ) )


    @property
    def comments( self ) -> dict:
        """
        Get the comments.
        """
        return self.metadata["comments"]
    
    @property
    def flags( self ) -> list:
        """
        Get the flags.
        """
        return self.metadata["flags"]



"""
The `records lookup` command can be used look up the latest comment of a given file.

Usage
-----

    >>> records lookup <filename>

    where ``<filename>`` is the file of interest.
    
"""



def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Look up a file's latest comment."
    parser = parent.add_parser( "lookup", description = descr, help = descr )
    parser.add_argument( "filename", nargs = "?", help = "The file to look up. If left blank the latest registry comment is looked up.", default = None )
    parser.set_defaults( func = lookup )

def lookup( args ):
    """
    The core function to look up the latest comment.
    """
    import filerecords.api as api
    import filerecords.api.utils as utils
    
    logger = utils.log()
    reg = api.Registry( "." )

    if not args.filename:

        last = reg.lookup_last()
        name = f"{reg.directory} (registry)"

    else:

        record = reg.get_record( args.filename )
        last = record.lookup_last() if record is not None else None
        name = f"{args.filename} ({', '.join(record.flags)})" if record is not None else args.filename

    if last:
        
        timestamp = list( last.keys() )[0]
        logger.debug( f"{last[timestamp]=}" )

        comment = last[timestamp]["comment"]
        user = last[timestamp]["user"]
        last = f"""{name}\n{comment}  |  {user} @ {timestamp.strftime("%Y-%m-%d %H:%M:%S")}"""
        print( last )


"""
The `records mv` command can be used to move / rename a file or directory in the registry.
This command by default also performs file moving. 

Usage
-----

    >>> records mv [-k] <filename>

    where ``<filename>`` is the path to the file to move / rename in the registry. 
    The ``-k`` option can be specified to prevent the file from being moved itself. In this case
    the file will remain unchanged while its records are adjusted.
"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Move / rename files or directories in the registry."
    parser = parent.add_parser( "mv", description = descr, help=descr )
    parser.add_argument( "current", help = "The file to move / rename." )
    parser.add_argument( "new", help = "The file's new path." )
    parser.add_argument( "-k", "--keep", help = "Keep the file itself and only adjust the records. By default the file or directory itself is also moved.", action = "store_true", default = False )
    parser.set_defaults( func = move )

def move( args ):
    """
    The core function to move / rename entries in the registry.
    """
    import filerecords.api as api
    # import filerecords.api.utils as utils

    # logger = utils.log()
    reg = api.Registry( "." )
    reg.move( args.current, args.new, keep_file = args.keep )



"""
Utility functions for filerecords.
"""

import logging
import subprocess
import yaml
from yaml.loader import SafeLoader
import os
import sys
import pandas as pd

import filerecords.api.settings as settings

def log( name : str = "filerecords", level : int = None, outfile : str = None ):
    """
    A logger that can be used to log records.

    Parameters
    ----------
    name : str
        The name of the logger.
    level : int
        The level of the logger.
    outfile : str
        The file to log to. If None, logs to stdout.
    """
    level = level if level else settings.log_level
    logger = logging.getLogger( name )
    logger.setLevel( level )

    if not logger.hasHandlers():

        if outfile is None:
            logger.addHandler( logging.StreamHandler( sys.stdout ) )
        else:
            logger.addHandler( logging.FileHandler( outfile ) )

    return logger

logger = log()

def make_new_registry( directory : str, perms : (int or str) = None ):
    """
    Make a new registry in a directory.

    Parameters
    ----------
    directory : str
        The directory to create the registry in.
    perms : int or str
        The permissions to set on the registry directory. This can be either the 
        numeric permissions (e.g. 755) or a string (e.g. "rwxr-xr-x").
        By default the permissions are inherited by the parent directory.
    """

    registry_dir = os.path.join( directory, settings.registry_dir )
    if os.path.exists( registry_dir ):
        raise FileExistsError( f"Registry already exists in {directory}" )
    
    os.makedirs( registry_dir )

    # add an INDEXFILE to the registry.
    indexfile = os.path.join( registry_dir, settings.indexfile )
    os.system( f"echo '{settings.indexfile_header}' > {indexfile}" )

    # add a METADATA file to the registry.
    _init_metafile( registry_dir )

    # set the permissions 
    if perms is None:
        perms = get_directory_perms(directory)
    os.system( f"chmod -R {perms} {registry_dir}" )

    logger.info( f"New registry created at {directory}" )

def get_directory_perms(directory):
    """
    Get the permissions of the parent directory.
    
    Parameters
    ----------
    directory : str
        The directory to get the permissions of the parent directory of.

    Returns
    -------
    int
        The permissions of the parent directory as numeric value.
    """
    perms = os.stat( directory ).st_mode
    perms = oct( perms )[-3:]
    perms = int( perms )
    return perms


def find_registry( directory :str ):
    """
    Find a registry associated with a source directory. 
    This will repeatedly search the upper level directories
    until a registry is found.

    Parameters
    ----------
    directory : str

    Returns
    -------
    str or None
        The path to the registry directory or None if no registry was found.
    """
    # paths = os.path.abspath( directory )
    paths = subprocess.check_output( f"cd {directory} ; pwd", shell = True ).decode().strip()
    while True:

        registry_dir = os.path.join( paths, settings.registry_dir )
        if os.path.exists( registry_dir ):
            logger.debug( f"Found registry in { os.path.dirname(registry_dir) }" )
            return registry_dir

        paths = os.path.dirname( paths )

        if paths == "/":
            return None

def get_indexfile( registry_dir : str ):
    """
    Get the indexfile of a registry.

    Parameters
    ----------
    registry_dir : str
        The path to the registry directory.
    
    Returns
    -------
    str
        The path to the registry's indexfile.
    """
    indexfile = os.path.join( registry_dir, settings.indexfile )

    if not os.path.exists( indexfile ):
        logger.critical( f"No indexfile found in {registry_dir}, broken registry?!" )
        return None

    return indexfile

def get_metafile( registry_dir : str ):
    """
    Get the metafile of a registry.

    Parameters
    ----------
    registry_dir : str
        The path to the registry directory.
    
    Returns
    -------
    str
        The path to the registry's metafile.
    """
    metafile = os.path.join( registry_dir, settings.registry_metafile )

    if not os.path.exists( metafile ):
        logger.critical( f"No metafile found in {registry_dir}, broken registry?!" )
        return None

    return metafile


def load_indexfile( filename : str ):
    """
    Load a registry indexfile.

    Parameters
    ----------
    filename : str
        The path to the registry indexfile.
    
    Returns
    -------
    pandas.DataFrame
        The contents of the registry indexfile.
    """
    df = pd.read_csv( filename, sep = "\t" )
    df.index = df["id"].values
    return df 

def load_yamlfile( filename : str ):
    """
    Load a yaml metadata file.

    Parameters
    ----------
    filename : str
        The path to the yaml file.
    
    Returns
    -------
    dict
        The contents of the yaml file.
    """
    with open( filename, "r" ) as f:
        contents = yaml.load( f, Loader = SafeLoader )

    return contents

def save_yamlfile( filename : str, contents : dict ):
    """
    Save a yaml metadata file.

    Parameters
    ----------
    filename : str
        The path to the yaml file.
    contents : dict
        The contents of the yaml file.
    """
    with open( filename, "w" ) as f:
        yaml.dump( contents, f )

def add_registry_to_gitignore():
    """
    Add the registry directory to a .gitignore file in the current directory.
    """
    gitignore = os.path.join( os.getcwd(), ".gitignore" )
    with open( gitignore, "a" ) as f:
        f.write( f"\n{settings.registry_dir}" )

def _init_metafile( registry_dir : str ):
    """
    Initialize a registry metadata file. 

    Parameters
    ----------
    registry_dir : str
        The registry directory.
    """
    contents = {
        "directory" : registry_dir,
        "comments" : {},
        "flags" : [],
        "groups" : {}
    }
    metafile = os.path.join( registry_dir, settings.registry_metafile )
    with open( metafile, "w" ) as f:
        yaml.dump( contents, f )

def _init_entryfile( registry_dir : str, id : str ):
    """
    Initialize a registry entry file. 

    Parameters
    ----------
    registry_dir : str
        The registry directory.
    id : str
        The id of the entry.
    """
    entryfile = os.path.join( registry_dir, id )
    with open( entryfile, "w" ) as f:
        yaml.dump( settings.entryfile_template, f )
    
    perms = get_directory_perms(registry_dir)
    os.system( f"chmod -R {perms} {entryfile}" )


"""
The `records comment` command can be used to add a comment to a specific file or the registry itself.

Usage
-----

    >>> records comment [-c <comment>] [-f <flags>] <filename>

    where ``<filename>`` is the path to the file to comment. If left blank the comment is added to the registry itself.
    ``-c <comment>``, adds a comment or description message.  ``-f <flags>``, adds any number of flags. These can include defined flag group labels.
"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Add comments to files or the registry itself."
    parser = parent.add_parser( "comment", description = descr, help = descr )
    parser.add_argument( "filename", nargs = "?", help = "The file to comment. If left blank the comments are applied to the registry itself", default = None )
    parser.add_argument( "-c", "--comment", help = "Add a comment or description.", default = None )
    parser.add_argument( "-f", "--flags", help = "Add flags.", nargs="+", default = None )
    parser.set_defaults( func = comment )

def comment( args ):
    """
    The core function to add comments.
    """
    import filerecords.api as api
    # from filerecords.api.utils import log

    # logger = log()

    reg = api.Registry( "." )

    # # the add and update functions are expecting to find a string
    # # if noly one value is present...
    # if args.flags and len(args.flags) == 1:
    #     args.flags = args.flags[0]

    if not args.filename:

        if args.comment:
            reg.add_comment( args.comment )

        if args.flags:
            reg.add_flags( args.flags )

        reg.save()
        
    else:

        record = reg.get_record( args.filename )
        if record is None:

            reg.add( args.filename, comment = args.comment, flags = args.flags )
            # logger.info( f"Added {args.filename} to the registry." )
            print( f"Added {args.filename} to the registry." )

        else:

            reg.update( args.filename, comment = args.comment, flags = args.flags )
            # logger.info( f"Updated {args.filename}'s records." )
            print( f"Updated {args.filename}'s records." )


"""
`FileRecord` is the basic object in which records within a registry are accessed.
This object can access and edit records for a specific file within the registry.

API Usage
=========

In the code API, a `Registry` will return a `FileRecord` object for a specific file when a record is accessed.
Each `FileRecord` requires a `Registry` object as parent to be initialized, and can then either access the registry
records via the internal id (happens when records are accessed via the `Registry` object) or via the file path which is then
searched for in the registry. It is suggested to access the records directly via the `Registry` object only.


Once a `FileRecord` has been obtained from the registry its comments and flags can be accessed using the `comments` and `flags` attributes.

.. code-block:: python

    from filerecords.api import Registry

    reg = Registry()

    # Access a record via the registry
    record = reg.get_record( "path/to/file" )

    # Access the comments
    comments = record.comments

    # Access the flags
    flags = record.flags
    

The `FileRecord` object can also be used to add new comments and flags to the record.

.. code-block:: python

    record.add_comment( "the new comment" )
    record.add_flag( ["flag1", "flag2"] )


When adding comments or flags it is important to save the record afterward. 
This will automatically also save the parent registry.

.. code-block:: python

    # Save the record (will also save the registry)
    record.save()


"""

from datetime import datetime
import uuid
import os

import filerecords.api.base as base
# import filerecords.api.registry as registry
import filerecords.api.utils as utils
import filerecords.api.settings as settings

logger = utils.log()

class FileRecord(base.BaseRecord):
    """
    This class represents a single file record entry.

    Parameters
    ----------
    registry : Registry
        The registry the file record is associated with.
    id : str 
        The unique identifier of the file record.
        If none is provided, a new id is created.
    filename : str 
        The filename of the file to record (if a new record is being created).
    """
    def __init__( self, registry, id : str = None, filename : str = None ):
        super().__init__()
        self.registry = registry
        self.filename = os.path.join( os.path.join( self.registry.directory, filename ) ) if filename else None


        if id is None:

            if filename is None:
                raise ValueError( "A filename must be provided when adding a new file record." )
            
            _init_new = True
            self.id = uuid.uuid4()

        else:

            _init_new = False
            self.id = id


        self.relpath = self._get_relpath()
        self.filename = self._get_filename()

        self.metafile = os.path.join( self.registry.registry_dir, str(self.id) )
        if _init_new:
            utils._init_entryfile( self.registry.registry_dir, str(self.id) )
        
        self.load()

        logger.debug( f"filename: {self.filename}" )


    def load( self ):
        """
        Loads the file records.

        Note
        ----
        This is done automatically during init if an existing file is specified.
        """
        super().load( self.metafile )

    def save( self ):
        """
        Save the file record.
        This will also save the the 
        registry state at the same time.
        """
        super().save()
        self.registry.save()

    def add_flags( self, flags : str or list ):
        """
        Add flags to the metadata.

        Note
        ----
        This will not automatically save the metadata,
        use the save() method to do so.

        Parameters
        ----------
        flags : str or list
            The flag(s) to add. 
            This can also be a defined flag-group label.
        """
        get_group_flags = lambda x: self.registry.groups[x] if x in self.registry.groups else [x]
        if isinstance( flags, list ):
            _flags = []
            for flag in flags:
                _flags += get_group_flags( flag )
            flags = _flags
            del _flags
        else:
            flags = get_group_flags( flags )
        super().add_flags( flags )
        self.registry.add_flags( flags )


    def to_markdown( self, comments_header : bool = True ):
        """
        Convert the metadata to a markdown representation.

        Parameters
        ----------
        comments_header : bool
            Add a header above the comments.
        """
        # the [3:] is to remove the ../ in the relpath beginning 
        # since every relpath starts at the base directory which is one
        # level above the registry directory.

        path = self.relpath[3:].replace( "_", "\_" )
        text = f"### {path}\n\n"

        if len( self.flags ) == 0:
            flags = "No flags"
        else:
            flags = '\n- '.join( self.flags )
        text += f"- {flags}\n\n"

        if comments_header:
            text += "#### Comments\n\n"

        if len( self.comments ) == 0:

            text += "No comments\n\n"

        else:

            for timestamp in self.comments:
                comment, user = list( self.comments[timestamp].values() )
                text += f"{settings.comment_format( comment, user, timestamp)}\n\n"
            
        return text

    def to_yaml( self, timestamp : bool = False, filename : str = None ):
        """
        Convert the source registry to a single YAML file.

        Parameters
        ----------
        timestamp : bool
            Add a timestamp in the markdown.
        filename : str (optional)
            The filename of the yaml file to create.
            If none is provided, no file is created.

        Returns
        -------
        dict
            The assembled dictionary of the registry.
        """
        
        _dict = dict( self.metadata )
        _dict["filename"] = self.filename
        _dict["relpath"] = self.relpath[3:] # the [3:] is to remove the ../ in front of every relpath
        if timestamp:
            _dict["timestamp"] = datetime.now().strftime( "%Y-%m-%d %H:%M:%S" )

        if filename is not None:
            utils.save_yamlfile( filename, _dict )

        return _dict


    def _get_relpath(self):
        """
        Make the path of the recorded file relative to the registry.
        """

        # logger.debug( f"Registry index: {self.registry.index}" ) 
        ids = self.registry.index.id.astype(str)

        # logger.debug( f"ids={ids}" )
        logger.debug( f"str(self.id)={str(self.id)}" )

        if str(self.id) in ids:
            return self.registry.index.loc[ids == str(self.id), "relpath"].values[0]
        
        else:

            relpath = os.path.relpath( self.filename, self.registry.registry_dir )

            if relpath in self.registry.index.relpath.values:
                directory = os.path.relpath( os.path.dirname(relpath), self.registry.directory ) 
                raise FileExistsError( f"File {self.filename} within {directory} already exists in the registry." )
        
            return relpath

    def _get_filename( self ):
        """
        Get the filename of the file record.
        """
        ids = self.registry.index.id.astype(str)

        # logger.debug( f"ids={ids}" )
        logger.debug( f"self.id={self.id}" )

        if str(self.id) in ids:

            logger.debug( self.registry.index )
            logger.debug( ids == str(self.id) )
            logger.debug( self.registry.index.loc[ids == str(self.id), "filename"].values[0] )

            return self.registry.index.loc[ids == str(self.id), "filename"].values[0]
        return os.path.basename( self.filename )

    def __repr__( self ):
        return f"{self.__class__.__name__}(id = {self.id}, filename = {self.filename})"


"""
The `records init` command can be used to initialize a new registry in the current directory.

Usage:

    >>> records init [-c <comment>] [-g <groupname> : <grouplabels>] [-g <groupname> : <grouplabels>]

    -c <comment> : Adds a comment or description to the registry that is initialized.
    -g <groupname> : Adds a label group to the registry. Label groups can be specified via '<name> : <label1> <label2>...' syntax. 
    Note, this option specifies a single group. It can be supplied multiple times to specify multiple groups in one go.

Flags and Flag groups
---------------------
Flags are used to label files and group them together. Flags are arbitrary labels so they can be specified on-the-run while commenting a file. Flag groups can be used to add multiple flags at a time to a file. 
Contrary to single flags, flag groups must be defined before they can be used. 

"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Initialize a new registry within the current directory."
    parser = parent.add_parser( "init", description = descr, help = descr )
    parser.add_argument( "-c", "--comment", help = "Add a comment or description to the registry", default = None )
    parser.add_argument( "-f", "--flags", help = "Add flags.", nargs="+", default = None)
    parser.add_argument( "-g", "--group", help = "Add flag groups to the registry. Flag groups can be specified via '<name> : <flag1> <flag2>...' syntax. Note, this option specifies a single group. It can be supplied multiple times to specify multiple groups in one go.", nargs="+", action = "append", default = None )
    parser.add_argument ("-i", "--gitignore", help = "Add the registry to .gitignore", action = "store_true" )
    parser.set_defaults( func = init )


def init( args ): 
    """
    The core function to initialize a new registry.
    """
    import filerecords.api as api
    import filerecords.cli.auxiliary as aux
    import filerecords.api.utils as utils

    logger = utils.log()

    reg = api.Registry( "." )

    if not reg._initialized and reg.base_has_registry():
        logger.info( "A registry already exists in this directory. Skipping initialization. Use the clear command to clear the registry." )
        return 

    # the init of the Registry should have
    # already taken care of this but just to be sure...
    elif not reg._initialized:
        reg.init()

    if args.comment:
        reg.add_comment( args.comment )
    
    if args.flags:
        reg.add_flags( args.flags )

    if args.group:
        aux._register_groups( reg, args.group )

    if args.gitignore:
        utils.add_registry_to_gitignore()

    reg.save()



"""
The `records rm` command can be used to remove a file or directory from the registry.
This command by default also performs file removal. 

Usage
-----

    >>> records rm [-k] <filename>

    where ``<filename>`` is the path to the file to remove from the registry. 
    The ``-k`` option can be specified to prevent the file from being removed itself. In this case
    the file will remain while its records are removed.
"""

def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Remove files from the registry."
    parser = parent.add_parser( "rm", description = descr, help=descr )
    parser.add_argument( "filename", help = "The file to remove." )
    parser.add_argument( "-k", "--keep", help = "Keep the file itself and only remove the records. By default the file or directory itself is also removed.", action = "store_true", default = False )
    parser.set_defaults( func = remove )

def remove( args ):
    """
    The core function to remove entries from the registry.
    """
    import filerecords.api as api
    # from filerecords.api.utils import log
        
    # logger = log()
    reg = api.Registry( "." )
    reg.remove( args.filename, keep_file = args.keep )
    # logger.info( f"Removed {args.filename} from the registry." )
    print( f"Removed {args.filename} from the registry." )


"""
The `records clear` command can be used clear all records of a registry.
This action is equivalent to destroying the old registry and creating a new one.

Usage:

    >>> records clear [-e <export_file>]

    To prevent loss of records, the records can be exported either to a yaml or a markdown file (or both) 
    before clearing. The ``-e`` option can be specified to export the records to a file. 
"""

def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Clear the registry."
    parser = parent.add_parser( "clear", description = descr, help = descr )
    parser.add_argument( "-e", "--export", help = "Export the registry before clearing. This will create a default 'registry-{timestamp}' file in either yaml or markdown format, or both, in the current directory.", choices = ["yaml", "md", "both"] )
    parser.add_argument( "-y", help = "Skip the confirmation prompt.", action = "store_true" )
    parser.set_defaults( func = clear )


def clear( args ): 
    """
    The core function to clear the registry.
    """
    import shutil
    import filerecords.api as api
    import filerecords.api.settings as settings
    import filerecords.api.utils as utils
    from datetime import datetime

    logger = utils.log()

    reg = api.Registry( "." )

    if args.export:
        filename = f"{settings.registry_export_name}-{datetime.now()}"
        man = api.Manifest( reg )

        if args.export == "yaml" or args.export == "both":
            man.to_yaml( filename + ".yaml" )
        if args.export == "md" or args.export == "both":
            man.to_markdown( filename + ".md" )
    
    if not args.y:

        confirmation = input( "This is a destructive action! Are you sure you wish to clear your records? [y|N]: ").lower()
        if not confirmation == "y":
            logger.info( "Aborting..." )
            return

    # remove the old registry
    shutil.rmtree( reg.registry_dir )

    # make a new registry
    reg.init()


"""
The `records flag` command can be used to add flags to a specific file or the registry itself, and to define flag groups.

Usage
-----

    >>> records flag [-f <flags>] [-g <groupname> : <grouplabels>] <filename>

    where ``<filename>`` is the path to the file to comment. If left blank the flag is added to the registry itself.
    ``-f <flags>``, adds any number of flags. These can include defined flag group labels.
    ``-g <groupname> : <grouplabels>``, will define a flag group in the registry. 
    ``-f`` and ``-g`` can be provided at the same time. 
"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Add flags to files or the registry itself (can also be done with comment), and define flag groups (this command only)."
    parser = parent.add_parser( "flag", description = descr, help = descr )
    parser.add_argument( "filename", nargs = "?", help = "The file to comment. If left blank the comments are applied to the registry itself", default = None )
    parser.add_argument( "-f", "--flags", help = "Add flags.", nargs="+", default = None )
    parser.add_argument( "-g", "--group", help = "Add flag groups to the registry. Flag groups can be specified via '<name> : <flag1> <flag2>...' syntax. Note, this option specifies a single group. It can be supplied multiple times to specify multiple groups in one go.", nargs="+", action = "append", default = None )

    parser.set_defaults( func = flag )

def flag( args ):
    """
    The core function to add comments.
    """
    import filerecords.api as api
    import filerecords.cli.auxiliary as aux
    # import filerecords.api.utils as utils

    # logger = utils.log()
    reg = api.Registry( "." )

    # add any new groups to the registry.
    if args.group:
        aux._register_groups( reg, args.group )

    # this bit is (almost) identical to the comment function

    if args.flags:

    #    if len(args.flags) == 1: 
    #     args.flags = args.flags[0]

        if not args.filename:

            if args.flags:
                reg.add_flags( args.flags )
            reg.save()

        else:

            record = reg.get_record( args.filename )
            if record is None:
                reg.add( args.filename, flags = args.flags )
            else:
                reg.update( args.filename, flags = args.flags )


import setuptools

with open( "README.md", "r" ) as fh:
    long_description = fh.read()

setuptools.setup(
    name="filerecords", 
    version="0.0.1",
    author="Noah H. Kleinschmidt",
    author_email="noah.kleinschmidt@students.unibe.ch",
    description="A command-line toolbox to keep file metadata in an organized and easily accessible way through comments and flags.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NoahHenrikKleinschmidt/filerecords",
    
    # the nuitka stuff is put on ice for the time being...
    # command_options={
    #   'nuitka': {
    #      # boolean option, e.g. if you cared for C compilation commands
    #      '--show-scons': True,
    #      # options without value, e.g. enforce using Clang
    #      '--clang': None,
    #      # options with single values, e.g. enable a plugin of Nuitka
    #      '--enable-plugin': "pyside2",
    #      # options with several values, e.g. avoiding including modules
    #      '--nofollow-import-to' : ["*.tests", "*.distutils", #],
    #      #'--noinclude-custom-mode' : [
    #                                     "matplotlib", 
    #                                     "IPython", 
    #                                     "bokeh", 
    #                                     "sphinx",
    #                                     "pytest",
    #                                     "tkinter",
    #                                     "sqlalchemy",
    #                                     "jupyter",
    #                                     "ipywidgets",
    #                                     "PIL",
    #                                     "PyQt5",
    #                                 ]
    #   }
    # },
    
    packages=setuptools.find_packages(),

    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],

    install_requires=[
        "pandas",
        "PyYAML",
        ],

    entry_points={
        "console_scripts": [ 
            "filerecords=filerecords.cli:setup",
            "records=filerecords.cli:setup",
        ]
    },
    
    python_requires='>=3.6',
)


"""
Auxiliary functions for the CLI.
"""

def _prep_groups( groups ):
    """
    Prepare the given flag groups for registry adding.
    This will make sure the given groupnames and associated flags are in a proper dictionary form.
    
    Parameters
    ----------
    groups : list of lists
        The groups to prepare.
    
    Returns
    -------
    dict
        The prepared groups. As { name : [flag1, flag2,...], another : [flag3,...], ... }
    """
    # first join all together ( and remove any whitespaces around the : )
    groups = [ " ".join(i).replace(" :", ":").replace(": ", ":" ) for i in groups ]
    # then separate the group name by the : 
    groups = [ i.split(":") for i in groups ]
    # now split the flags into lists and assemble all again
    groups = { i[0] : i[1].split(" ") for i in groups }
    return groups

def _register_groups( registry, groups ):
    """
    Register the given groups in the registry.

    Parameters
    ----------
    registry : Registry
        The registry to register the groups in.
    groups : dict
        The groups to register. As { name : [flag1, flag2,...], another : [flag3,...], ... }
    """
    groups = _prep_groups( groups )
    for name, flags in groups.items():
        registry.add_group( name, flags )
    registry.save()


"""
This is the main command line interface for `filerecords`. 
"""

import argparse
import filerecords.cli.init as init
import filerecords.cli.comment as comment
import filerecords.cli.flag as flag
import filerecords.cli.move as move
import filerecords.cli.remove as remove
import filerecords.cli.undo as undo
import filerecords.cli.list as list
import filerecords.cli.list_local as list_local
import filerecords.cli.lookup as lookup
import filerecords.cli.read as read 
import filerecords.cli.clear as clear
import filerecords.cli.export as export
import filerecords.cli.destroy as destroy


def setup():
    """
    Setup the command line interface.
    """
    descr = "filerecords – a command line tool for storing file metadata in a structured way."
    parser = argparse.ArgumentParser( description = descr )
    parser.add_argument("-v", "--version", action="store_true", help="Show version")
    subparsers = parser.add_subparsers(help="Available commands")

    # Add the sub-commands here.
    init.setup(subparsers)
    comment.setup(subparsers)
    flag.setup(subparsers)
    undo.setup(subparsers)
    lookup.setup(subparsers)
    read.setup(subparsers)
    move.setup(subparsers)
    remove.setup(subparsers)
    list.setup(subparsers)
    list_local.setup(subparsers)
    export.setup(subparsers)
    clear.setup(subparsers)
    destroy.setup(subparsers)
    

    args = parser.parse_args()

    if args.version:
        print("filerecords version 0.0.1")

    else:
        args.func( args )

if __name__ == "__main__":
    setup()


"""
The `records read` command can be used read the entire records of a given file.

Usage
-----

    >>> records read <filename>

    where ``<filename>`` is the file of interest.
    
"""


def setup( parent ):
    """
    Set up the CLI
    """
    descr = "Read a file's records."
    parser = parent.add_parser( "read", description = descr, help = descr )
    parser.add_argument( "filename", nargs = "?", help = "The file whose records to read. If left blank the registry's own records are read.", default = None )
    parser.set_defaults( func = read )

def read( args ):
    """
    The core function to read records.
    """
    import subprocess
    import filerecords.api as api
    # import filerecords.api.utils as utils
    
    # logger = utils.log()
    reg = api.Registry( "." )

    if not args.filename:

        records = reg.to_markdown( include_records = False )

    else:

        record = reg.get_record( args.filename )
        records = record.to_markdown() if record is not None else None

    if records:
        
        # we try to show the markdown file 
        # with glow rather than printing it out blankly...
        try:
            
            # first check if we have glow installed
            out = subprocess.run( "glow -h", shell=True, stdout = subprocess.PIPE, stderr = subprocess.PIPE )
            if out.returncode != 0:
                raise RuntimeError

            # if we have it installed, call glow to render the markdown
            subprocess.run( f"echo '{records}' | glow -" , shell = True )
           
        # if we don't have glow, just print the markdown normally...
        except RuntimeError:

            print( records )